import json
from fractions import Fraction

from ..grades import PartGrade
from ..submission import BrokenSubmissionError
from ..utils import PIPE, TimeoutExpired, run_process
from . import GraderInterface, Part

"""
Grade a homework using a LC3tools grader on the new testing framework
(API_VER 2110)
"""


class LC3ToolsTest(Part):
    __slots__ = ("test",)

    def __init__(self, test):
        self.test = test

    def description(self):
        return self.test

    def grade(self, result):
        if result is None:
            msg = (
                "Results for test not found. Check if there were any "
                "assembler erros when assembling this file on lc3tools. If not"
                ", report this as an autograder error to the instructors."
            )
            return PartGrade(score=Fraction(0), log=msg)

        # display potentially helpful output if test failed
        partialFailures = len(result["partialFailures"])
        log = result["output"] if partialFailures else ""

        log += "\n".join(
            "--{}: {}".format(failure["displayName"], failure["message"])
            for failure in result["partialFailures"]
        )

        if partialFailures < result["failed"]:
            log += "\n[omitted {} more failures]".format(
                result["failed"] - partialFailures
            )

        score = Fraction(result["total"] - result["failed"], result["total"])
        return PartGrade(score=score, log=log)


class LC3ToolsGrader(GraderInterface):
    """
    Run a LC3Tools test executable, which can output json with the new
    testing framework (much more reliable than regex). Pretty heavily inspired
    by the CircuitSim and PyLC3 graders.
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, test_file, asm_file, timeout=None):
        self.test_file = test_file
        self.asm_file = asm_file

        self.timeout = self.DEFAULT_TIMEOUT if timeout is None else timeout

    def list_prerequisites(self):
        # test executables are already built
        return []

    def part_from_config_dict(self, config_dict):
        return LC3ToolsTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        cmdline = [
            "./" + self.test_file,
            self.asm_file,
            "--json-output",
            "--asm-print-level=3",
        ]
        try:
            # Do not mix stderr into stdout because sometimes our friend
            # Roi printStackTrace()s or System.err.println()s, and that
            # will mess up JSON parsing
            process = run_process(
                cmdline,
                cwd=path,
                timeout=self.timeout,
                stdout=PIPE,
                stderr=PIPE,
                input="",
            )
        except TimeoutExpired:
            raise BrokenSubmissionError(
                "timeout of {} seconds expired for grader".format(self.timeout)
            )

        if process.returncode != 0:
            raise BrokenSubmissionError(
                "grader command exited with nonzero exit code {}".format(
                    process.returncode
                ),
                verbose="\n".join(
                    output_stream.decode()
                    for output_stream in [process.stderr, process.stdout]
                    if output_stream
                ),
            )

        results = json.loads(process.stdout.decode(errors='backslashreplace'))

        if "error" in results:
            raise BrokenSubmissionError(results["error"])

        method_results = {res["testName"]: res for res in results["tests"]}
        return [part.grade(method_results.get(part.test)) for part in parts]



import os

import xml.etree.ElementTree
from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import Part, GraderInterface


class EnsembleTest(Part):
    __slots__ = ('name')

    def __init__(self, name):
        self.name = name

    def description(self):
        return self.name

    def grade(self, result):
        if result is None:
            msg = ('Results for test not found. Check if there were any'
                   ' internal errors reported. If not, report this as an'
                   ' autograder error to your instructors.')
            return PartGrade(score=0, log=msg)

        failure = result.find('failure')

        if failure is not None:
            log = "\n".join([failure.get('message'), failure.text])
            return PartGrade(score=0, log=log)

        return PartGrade(score=1, log='')

class EnsembleGrader(GraderInterface):
    def __init__(self, test_file, timeout=30):
        self.test_file = test_file
        self.timeout = timeout

    def list_prerequisites(self):
        return []

    def part_from_config_dict(self, config_dict):
        return EnsembleTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        command = ['pytest', self.test_file, '--junitxml', 'report.xml']

        try:
            process = run_process(
                command,
                cwd=path,
                timeout=self.timeout,
                stdout=PIPE,
                stderr=STDOUT,
                input=''
            )
        except TimeoutExpired:
            raise BrokenSubmissionError(
                "timeout of {} seconds expired for grader"
                .format(self.timeout)
            )

        if process.returncode != 0 and process.returncode != 1:
            raise BrokenSubmissionError(
                'grader command exited with exit code {}'
                .format(process.returncode),
                verbose=process.stdout.decode(errors='backslashreplace') if process.stdout else None
            )

        report_xml_path = os.path.join(path, 'report.xml')
        report_xml = xml.etree.ElementTree.parse(report_xml_path)

        return [part.grade(report_xml.find(f'.//testcase[@name=\'{part.name}\']')) for part in parts]



import re

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade, Fraction
from . import Part, ThreadedGrader

class CriterionTest(Part):
    __slots__ = ('name', 'suite', 'test')

    def __init__(self, name, suite, test, valgrind_deduction="1/2"):
        self.name = name
        self.suite = suite
        self.test = test
        self.valgrind_deduction = Fraction(valgrind_deduction)
    
    def description(self):
        return self.name
    
    def grade(self, path, grader):
        flags = [
            "--timeout", str(grader.single_timeout),
            f"--filter={self.suite}/{self.test}",
        ]
        command = ['./tests', *flags]
        if grader.valgrind_cmd is not None:
            valgrind_cmd = grader.valgrind_cmd + flags
        else:
            valgrind_cmd = None

        try:
            process = run_process(valgrind_cmd if valgrind_cmd else command, cwd=path, stdout=PIPE, stderr=STDOUT, timeout=grader.total_timeout)
        except TimeoutExpired:
            raise BrokenSubmissionError(f'grader timed out after {grader.total_timeout} seconds')
        
        result = process.stdout.decode(errors='backslashreplace')

        if result is None:
            msg = ('Results for test not found. Check if there were any'
                   ' internal errors reported. If not, report this as an'
                   ' autograder error to your instructors.')
            return PartGrade(score=0, log=msg)
        
        pattern = r"Tested:\s*(\d+)\s*\|\s*Passing:\s*(\d+)"

        matches = re.search(pattern, result)

        if not matches:
            return PartGrade(score=0, log="Could not parse autograder output")
        
        total, passing = int(matches.group(1)), int(matches.group(2))

        if total == 0:
            return PartGrade(score=0, log="No test cases were found")

        if valgrind_cmd is not None and re.findall(r"^==\d+==.*$", result, re.MULTILINE):
            return PartGrade(passing / total * (1 - self.valgrind_deduction), log=result)

        if total == passing:
            return PartGrade(score=1, log="")
        
        log_line_pattern = r"^\s*\[(?:----|FAIL)\].*"
        matches = re.findall(log_line_pattern, result, re.MULTILINE)

        return PartGrade(score=passing / total, log="\n".join(matches))

class CriterionGrader(ThreadedGrader):
    def __init__(self, valgrind_cmd: "str | None" = None, total_timeout: float = 60, single_timeout: float = 3):
        super(CriterionGrader, self).__init__(None)
        if valgrind_cmd is not None:
            self.valgrind_cmd = valgrind_cmd.split(" ")
        else:
            self.valgrind_cmd = None
        self.single_timeout = float(single_timeout)
        self.total_timeout = float(total_timeout)

    def list_prerequisites(self):
        return []

    def part_from_config_dict(self, config_dict):
        return CriterionTest.from_config_dict(config_dict)
    
    def grade_part(self, part, path, submission):
        return part.grade(path, self)

    def grade(self, submission, path, parts):
        command = ['make', 'tests']

        try:
            process = run_process(
                command,
                cwd=path,
                timeout=self.total_timeout,
                stdout=PIPE,
                stderr=STDOUT,
                input=''
            )
        except TimeoutExpired:
            raise BrokenSubmissionError(
                "timeout of {} seconds expired for grader"
                .format(self.total_timeout)
            )

        if process.returncode != 0 and process.returncode != 1:
            raise BrokenSubmissionError(
                'grader command exited with exit code {}\n'
                .format(process.returncode),
                verbose=process.stdout.decode(errors='backslashreplace') if process.stdout else None
            )

        return super(CriterionGrader, self).grade(submission, path, parts)